@@ -1,13 +1,15 @@
-__version__ = "0.0.7"
+__version__ = "0.0.8"
 
 from .core import (
+    is_np,
     broadcast_batch,
     to_hom,
     homogenize,
+    look_at_torch,
+    look_at_np,
     compose_rt,
     to_44,
     to_34,
-    look_at_np,
     to_np,
     to_numpy,
     to_torch,
@@ -18,9 +20,12 @@ from .core import (
     rotz,
     map_range,
     create_random_cameras_on_unit_sphere,
+    random_vectors_on_sphere,
     opengl_c2w_to_opencv_c2w,
     opencv_c2w_to_opengl_c2w,
+    opencv_intrinsics_from_opengl_project,
     opengl_project_from_opencv_intrinsics,
+    perspective_projection,
     vec2skew,
     batch_vec2skew,
     rotvec2mat,
@@ -74,6 +79,7 @@ from .image import (
     draw_text_on_image,
     draw_gizmo_on_image,
     merge_figures_with_line,
+    generate_checkerboard,
     generate_voronoi_diagram,
     generate_gray_gradient,
     generate_dot_pattern,
@@ -84,8 +90,13 @@ from .image import (
     interpolate_multi_channel,
     image_grid,
     resize_images_naive,
+    adjust_contrast_brightness,
     change_brightness,
-    change_contrast
+    linear_to_srgb,
+    srgb_to_linear,
+    pad_image_to_res,
+    mask_regions,
+    crop_center
 )
 
 from .video import (
