@@ -2,24 +2,35 @@ import torch
 import numpy as np
 from PIL import Image
 
+def is_np(x):
+    """
+    checks if x is a numpy array or torch tensor (will raise an error if x is neither)
+    :param x: object to check
+    :return: True if x is a numpy array, False if x is a torch tensor
+    """
+    if type(x) == np.ndarray:
+        return True
+    elif type(x) == torch.Tensor:
+        return False
+    else:
+        raise ValueError("input must be torch.Tensor or np.ndarray")
+
 def to_hom(x):
     """
     converts a vector to homogeneous coordinates
     :param x: nxc numpy array
     :return: nxc+1 numpy array
     """
-    if type(x) == torch.Tensor:
-        if x.ndim == 1:
-            return torch.cat((x, torch.ones(1, device=x.device)))
-        else:
-            return torch.cat((x, torch.ones(x.shape[0], 1, device=x.device)), dim=-1)
-    elif type(x) == np.ndarray:
+    if is_np(x):
         if x.ndim == 1:
             return np.concatenate((x, np.array([1], dtype=x.dtype)))
         else:
             return np.concatenate((x, np.ones((x.shape[0], 1), dtype=x.dtype)), axis=-1)
     else:
-        raise ValueError("x must be torch.Tensor or np.ndarray")
+        if x.ndim == 1:
+            return torch.cat((x, torch.ones(1, device=x.device)))
+        else:
+            return torch.cat((x, torch.ones(x.shape[0], 1, device=x.device)), dim=-1)
 
 def homogenize(x, keepdim=False):
     """
@@ -33,10 +44,11 @@ def homogenize(x, keepdim=False):
     return x
 
 def normalize(x, eps=1e-7):
-    if type(x) == torch.Tensor:
-        return x / (torch.norm(x, dim=-1, keepdim=True) + eps)
-    elif type(x) == np.ndarray:
+
+    if is_np(x):
         return x / (np.linalg.norm(x, axis=-1, keepdims=True) + eps)
+    else:
+        return x / (torch.norm(x, dim=-1, keepdim=True) + eps)
 
 
 def broadcast_batch(*args):
@@ -89,15 +101,13 @@ def to_44(mat):
         return mat
     if mat.shape[-2:] != (3, 4):
         raise ValueError("mat must be 3x4")
-    if type(mat) == torch.Tensor:
-        to_cat = torch.zeros((*mat.shape[:-2], 1, 4), dtype=mat.dtype, device=mat.device)
-        to_cat[..., -1] = 1
-        new_mat = torch.cat((mat, to_cat), dim=-2)
-    elif type(mat) == np.ndarray:
+    if is_np(mat):
         to_cat = np.broadcast_to(np.array([0, 0, 0, 1]), (*mat.shape[:-2], 1, 4))
         new_mat = np.concatenate((mat, to_cat), axis=-2)
     else:
-        raise ValueError("mat must be torch.Tensor or np.ndarray")
+        to_cat = torch.zeros((*mat.shape[:-2], 1, 4), dtype=mat.dtype, device=mat.device)
+        to_cat[..., -1] = 1
+        new_mat = torch.cat((mat, to_cat), dim=-2)
     return new_mat
 
 def to_34(mat: np.array):
@@ -115,59 +125,56 @@ def to_34(mat: np.array):
             raise ValueError("mat must be 4x4")
         return mat[:-1, :]
 
-def look_at_np(from_, to_, up_, openGL=False):
+def look_at_np(eye, at, up, opengl=False):
     """
     returns a batch of look_at transforms 4x4 (camera->world, the inverse of a ModelView matrix)
     will broadcast upon batch dimension if necessary.
-    :param from_: n x 3 from vectors in world space
-    :param to_: n x 3 at vector in world space
-    :param up_: n x 3 up vector in world space
-    :param is_openGL: if True, output will be in opengl coordinates (z backward, y up) otherwise (z forward, y down)
+    :param eye: n x 3 from vectors in world space
+    :param at: n x 3 at vector in world space
+    :param up: n x 3 up vector in world space
+    :param opengl: if True, output will be in OpenGL coordinates (z backward, y up) otherwise (z forward, y down)
     :return: n x 4 x 4 transformation matrices (camera->world, the inverse of a ModelView matrix)
     """
-    from_, to_, up_ = broadcast_batch(from_, to_, up_)
-    forward = to_ - from_
+    eye, at, up = broadcast_batch(eye, at, up)
+    forward = at - eye
     forward = forward / np.linalg.norm(forward, axis=-1, keepdims=True)
-    right = np.cross(forward, up_)
+    right = np.cross(forward, up)
     right = right / np.linalg.norm(right, axis=-1, keepdims=True)
     up = np.cross(forward, right)
     up = up / np.linalg.norm(up, axis=-1, keepdims=True)
     rot = np.concatenate((right[..., None], up[..., None], forward[..., None]), axis=-1)
-    c2w = np.concatenate((rot, from_[..., None]), axis=-1)
-    c2w = to_44(c2w)
-    if openGL:
-        c2w[:, :, 1] *= -1
-        c2w[:, :, 2] *= -1
+    c2w = compose_rt(rot, eye, square=True)
+    if opengl:
+        c2w = opencv_c2w_to_opengl_c2w(c2w)
     return c2w
 
 def look_at_torch(
         eye:torch.Tensor, #3
         at:torch.Tensor, #3
         up:torch.Tensor, #3
-        device:torch.device,
-        openGL:bool=False
+        opengl:bool=False
     ) -> torch.Tensor: #4,4
     """
-    creates a lookat transform matrix (OpenCV convention)
+    creates a lookat transform matrix
     :param eye: where the camera is
     :param at: where the camera is looking
     :param up: the up vector of world space
-    :param device: the device to put the matrix on
+    :param opengl: if True, output will be in OpenGL coordinates (z backward, y up) otherwise (z forward, y down)
     :return: 4x4 lookat transform matrix
     """
-    if openGL:
-        z = (eye - at).type(torch.float32).to(device)
-    else:
-        z = (at - eye).type(torch.float32).to(device)
+    # todo batch support
+    z = (at - eye).type(torch.float32)
     z /= torch.norm(z)
-    x = torch.cross(up, z).type(torch.float32).to(device)
+    x = torch.cross(z, up).type(torch.float32)
     x /= torch.norm(x)
-    y = torch.cross(z, x).type(torch.float32).to(device)
+    y = torch.cross(z, x).type(torch.float32)
     y /= torch.norm(y)
-    T = torch.eye(4, device=device)
-    T[:3,:3] = torch.stack([x,y,z],dim=1)
-    T[:3,3] = eye
-    return T
+    c2w = torch.eye(4, device=z.device)
+    c2w[:3,:3] = torch.stack([x,y,z],dim=1)
+    c2w[:3,3] = eye
+    if opengl:
+        c2w = opencv_c2w_to_opengl_c2w(c2w)
+    return c2w
 
 def orthographic_projection(l, r, b, t, n, f):
     """
@@ -233,13 +240,13 @@ def frustum_projection(x0, x1, y0, y1, z0, z1):
 
 def opengl_project_from_opencv_intrinsics(opencv_intrinsics, width, height, near=0.1, far=100.0):
     """
-    given a matrix K from opencv, returns the corresponding projection matrix for opengl ("Eye/Camera/View space -> Clip Space")
-    :param opencv_project: 3x3 projection matrix from opencv
+    given a matrix K from opencv, returns the corresponding projection matrix for OpenGL ("Eye/Camera/View space -> Clip Space")
+    :param opencv_intrinsics: 3x3 intrinsics matrix from opencv
     :param width: width of the image
     :param height: height of the image
     :param near: near plane
     :param far: far plane
-    :return: 4x4 projection matrix for opengl (note: column major)
+    :return: 4x4 projection matrix for OpenGL (note: column major)
     """
     fx = opencv_intrinsics[0, 0]
     fy = opencv_intrinsics[1, 1]
@@ -251,6 +258,24 @@ def opengl_project_from_opencv_intrinsics(opencv_intrinsics, width, height, near
                            [0.0, 0.0, -1.0, 0.0]])
     return opengl_mtx
 
+def opencv_intrinsics_from_opengl_project(opengl_project, width, height):
+    """
+    given a projection matrix from OpenGL ("Eye/Camera/View space -> Clip Space"), returns a matrix K for opencv
+    :param opengl_project: 4x4 projection matrix from OpenGL
+    :param width: width of the image
+    :param height: height of the image
+    :return: 3x3 intrinsics matrix for opencv
+    note: assumes camera center is middle of image
+    """
+    fx = opengl_project[0, 0] * width / 2
+    fy = opengl_project[1, 1] * height / 2
+    cx = width / 2
+    cy = height / 2
+    opencv_mtx = np.array([[fx, 0.0, cx],
+                           [0.0, fy, cy],
+                           [0.0, 0.0, 1]])
+    return opencv_mtx
+
 def opengl_c2w_to_opencv_c2w(opengl_transforms):
     """
     given a modelview matrix (World space->Eye/Camera/View space) where z is backward and y is up,
@@ -270,30 +295,40 @@ def opengl_c2w_to_opencv_c2w(opengl_transforms):
 
 def opencv_c2w_to_opengl_c2w(opencv_transform):
     """
-    converts coordinates of "vision" (opencv) to opengl coordinates by flipping y and z axes
+    converts coordinates of "vision" (opencv) to OpenGL coordinates by flipping y and z axes
     """
     return opengl_c2w_to_opencv_c2w(opencv_transform)
 
-def create_random_cameras_on_unit_sphere(n, r, device="cuda"):
+def create_random_cameras_on_unit_sphere(n_cams, radius, normal=None, opengl=False, device="cpu"):
     """
     creates a batch of world2view ("ModelView" matrix) and view2clip ("Projection" matrix) transforms on a unit sphere looking at the center
-    :param n: number of cameras
-    :param r: radius of the sphere
+    :param n_cams: number of cameras
+    :param radius: radius of the sphere
+    :param normal: if provided, only the hemisphere in the direction of the normal is sampled
+    :param opengl: if True, the coordinate system is converted to openGL convention (z backward, y up)
     :param device: device to put the tensors on
-    :return: world2view, view2clip
+    :return: world2view (world2cam), view2clip (requires further processing if opengl=False)
     """
-    locs = torch.randn((n, 3), device=device)
+    locs = torch.randn((n_cams, 3), device=device)
     locs = torch.nn.functional.normalize(locs, dim=1, eps=1e-6)
-    locs = locs * r
-    matrices = torch.empty((n, 4, 4), dtype=torch.float32, device=device)
+    if normal is not None:
+        if normal.ndim == 1:
+            normal = normal[None, :]
+        normal = torch.nn.functional.normalize(normal, dim=-1, eps=1e-6)
+        dot_product = (locs[:, None, :] @ normal[:, :, None]).squeeze()
+        locs[dot_product < 0] *= -1
+    locs = locs * radius
+    matrices = torch.empty((n_cams, 4, 4), dtype=torch.float32, device=device)
     for i in range(len(locs)):
         matrices[i] = look_at_torch(locs[i],
-                                       torch.zeros(3, dtype=torch.float32, device=device),
-                                       torch.tensor([0.,1.,0.], device=device),
-                                       device=device)
+                                    torch.zeros(3, dtype=torch.float32, device=device),
+                                    torch.tensor([0.,0.,1.], device=device),
+                                    opengl=opengl)
     v2w = matrices  # c2w
     w2v = torch.inverse(v2w)
-    v2c = torch.tensor(perspective_projection(), dtype=torch.float32, device=device)
+    v2c = perspective_projection()
+    v2c = np.tile(v2c[None, :], (n_cams, 1, 1))
+    v2c = torch.tensor(v2c, dtype=torch.float32, device=device)
     return w2v, v2c
 
 def to_np(arr: torch.Tensor):
@@ -331,29 +366,61 @@ def to_torch(arr: np.array, device="cpu", dtype=None):
     else:
         return torch.tensor(arr, dtype=dtype, device=device)
 
-def to_8b(x: np.array, clip=True):
+def to_8b(x, clip=True):
     """
-    convert a numpy (float, double) array to 8 bit
+    convert an array (float, double) array to 8 bit
+    :param x: array
+    :param clip: if True, clips values to [0,1]
+    :return: 8 bit array
     """
-    if x.dtype == np.float32 or x.dtype == np.float64:
-        if clip:
-            x = np.clip(x, 0, 1)
-        return (255 * x).astype(np.uint8)
-    elif x.dtype == np.uint8:
-        return x
-
-def to_float(x: np.array, clip=True):
-    """
-    convert a numpy (8bit) array to float
-    """
-    if x.dtype == np.uint8:
-        return x.astype(np.float32) / 255
-    elif x.dtype == np.float32:
-        if clip:
-            x = np.clip(x, 0, 1)
-        return x
+    if is_np(x):
+        if x.dtype == np.float32 or x.dtype == np.float64:
+            if clip:
+                x = np.clip(x, 0, 1)
+            return (255 * x).round().astype(np.uint8)
+        elif x.dtype == bool:
+            return x.astype(np.uint8) * 255
+        elif x.dtype == np.uint8:
+            return x
     else:
-        raise ValueError("unsupported dtype")
+        if x.dtype == torch.float32 or x.dtype == torch.float64:
+            if clip:
+                x = torch.clamp(x, 0, 1)
+            return (255 * x).round().type(torch.uint8)
+        elif x.dtype == torch.bool:
+            return x.type(torch.uint8) * 255
+        elif x.dtype == torch.uint8:
+            return x
+
+def to_float(x, clip=True):
+    """
+    convert a 8bit or bool array to float
+    :param x: array
+    :param clip: if True, clips values to [0,1]
+    :return: float array
+    """
+    if is_np(x):
+        if x.dtype == np.uint8:
+            return x.astype(np.float32) / 255
+        elif x.dtype == np.float32:
+            if clip:
+                x = np.clip(x, 0, 1)
+            return x
+        elif x.dtype == bool:
+            return x.astype(np.float32)
+        else:
+            raise ValueError("unsupported dtype")
+    else:
+        if x.dtype == torch.uint8:
+            return x.to(torch.float32) / 255
+        elif x.dtype == torch.float32:
+            if clip:
+                x = torch.clamp(x, 0, 1)
+            return x
+        elif x.dtype == torch.bool:
+            return x.to(torch.float32)
+        else:
+            raise ValueError("unsupported dtype")
 
 def to_PIL(x: np.array):
     """
@@ -392,13 +459,19 @@ def scale(s):
     mat = np.diag(s)
     return to_44(mat)
 
-def sincos(a):
-    a = np.deg2rad(a)
+def sincos(a, degrees=True):
+    """
+    sin and cos of an angle
+    :param a: angle
+    :param degrees: if True, a is in degrees
+    """
+    if degrees:
+        a = np.deg2rad(a)
     return np.sin(a), np.cos(a)
 
 def rotate(a, r):
     """
-    creates a rotation matrix from an angle and an axis
+    creates a rotation matrix from an angle a and an axis of rotation r
     """
     s, c = sincos(a)
     r = normalize(r)
@@ -409,31 +482,37 @@ def rotate(a, r):
                       [x*z*nc - y*s, y*z*nc + x*s, z*z*nc +   c, 0],
                       [           0,            0,            0, 1]])
 
-def rotx(a):
+def rotx(a, degrees=True):
     """
-    creates a rotation matrix around the x axis
+    creates a homogeneous 3D rotation matrix around the x axis
+    a: angle
+    degrees: if True, a is in degrees, else radians
     """
-    s, c = sincos(a)
+    s, c = sincos(a, degrees)
     return np.array([[1,0,0,0],
                       [0,c,-s,0],
                       [0,s,c,0],
                       [0,0,0,1]])
 
-def roty(a):
+def roty(a, degrees=True):
     """
-    creates a rotation matrix around the y axis
+    creates a homogeneous 3D rotation matrix around the y axis
+    a: angle
+    degrees: if True, a is in degrees, else radians
     """
-    s, c = sincos(a)
+    s, c = sincos(a, degrees)
     return np.array([[c,0,s,0],
                       [0,1,0,0],
                       [-s,0,c,0],
                       [0,0,0,1]])
 
-def rotz(a):
+def rotz(a, degrees=True):
     """
-    creates a rotation matrix around the z axis
+    creates a homogeneous 3D rotation matrix around the z axis
+    a: angle
+    degrees: if True, a is in degrees, else radians
     """
-    s, c = sincos(a)
+    s, c = sincos(a, degrees)
     return np.array([[c,-s,0,0],
                       [s,c,0,0],
                       [0,0,1,0],
@@ -505,7 +584,7 @@ def mat2rotvec(r: torch.Tensor):
 
 def random_qvec(n: int):
     """
-    Generate random quaternions representing rotations
+    generate random quaternions representing rotations
     :param n: Number of quaternions in a batch to return.
     :return: Quaternions as tensor of shape (N, 4).
     """
@@ -515,6 +594,25 @@ def random_qvec(n: int):
     o = o / denom
     return o
 
+def random_vectors_on_sphere(n, normal=None, device="cpu"):
+    """
+    create a batch of uniformly distributed random unit vectors on a sphere
+    note: if normal is provided, returns random unit vectors on the hemisphere around the normal, but isn't uniform anymore.
+    :param n: number of vectors
+    :param normal: normals to orient the hemisphere (,3) or (n,3)
+    :param device: device to put the tensors on
+    :return: tensor of shape (n, 3)
+    """
+    locs = torch.randn((n, 3), device=device)
+    locs = torch.nn.functional.normalize(locs, dim=1, eps=1e-6)
+    if normal is not None:
+        if normal.ndim == 1:
+            normal = normal[None, :]
+        normal = torch.nn.functional.normalize(normal, dim=-1, eps=1e-6)
+        dot_product = (locs[:, None, :] @ normal[:, :, None]).squeeze()
+        locs[dot_product < 0] *= -1
+    return locs
+
 def qvec2mat(qvec):
     """
     Converts a quaternion to a rotation matrix
@@ -535,14 +633,12 @@ def batch_qvec2mat(qvecs):
         qvecs = qvecs[None, :]
     if qvecs.ndim > 2:
         raise ValueError("quaternions must be of shape (..., 4)")
-    if type(qvecs) == torch.Tensor:
-        r, i, j, k = torch.unbind(qvecs, -1)
-        stacking_func = torch.stack
-    elif type(qvecs) == np.ndarray:
+    if is_np(qvecs):
         r, i, j, k = [x[0] for x in np.split(qvecs, 4, -1)]
         stacking_func = np.stack
     else:
-        raise ValueError("quaternions must be of type torch.Tensor or np.ndarray")
+        r, i, j, k = torch.unbind(qvecs, -1)
+        stacking_func = torch.stack
     two_s = 2.0 / (qvecs * qvecs).sum(-1)
     o = stacking_func([1 - two_s * (j * j + k * k),
                        two_s * (i * j - k * r),
