@@ -68,7 +68,7 @@ def compute_v_hats(v, valid_pairs, Qv):
     v_hats[~mask] = np.mean(v[valid_pairs[~mask]], axis=1)  # if Q_hats is singular, use the mean of the two vertices
     return v_hats
 
-def qslim(v: np.ndarray, f: np.ndarray, budget: int):
+def qem(v: np.ndarray, f: np.ndarray, budget: int):
     """
     A slightly naive (and slow) implementation of "Surface Simplification Using Quadric Error Metrics", 1997
     """
@@ -149,4 +149,27 @@ def distribute_vector_field(num_vertices, f, per_face_vector_field):
     incident_face_vectors = torch.zeros([num_vertices,per_face_vector_field.shape[-1]], device=device)
     incident_face_vectors = torch.index_add(incident_face_vectors, dim=0, index=f_unrolled,
                                             source=normals_repeated_per_face)
-    return incident_face_vectors
\ No newline at end of file
+    return incident_face_vectors
+
+def rigid_transform_3d(A, B):
+    """
+    finds best rigid transformation between pc a and pc b (in terms of rmse) based on "Least-Squares Rigid Motion Using SVD"
+    :param A: point cloud a
+    :param B: point cloud b
+    :return: R, t such that B = R @ A + t
+    """
+    centroid_A = np.mean(A, axis=0)
+    centroid_B = np.mean(B, axis=0)
+
+    A_mean = A - centroid_A
+    B_mean = B - centroid_B
+
+    H = A_mean.T @ B_mean
+    U, S, Vt = np.linalg.svd(H)
+
+    flip = np.linalg.det(Vt.T @ U.T)
+    ones = np.identity(len(Vt))
+    ones[-1, -1] = flip
+    R = Vt.T @ ones @ U.T
+    t = centroid_B - R @ centroid_A
+    return R, t
\ No newline at end of file
