@@ -1,7 +1,8 @@
 import numpy as np
+import torch
 from PIL import Image, ImageDraw, ImageFont
 from scipy import interpolate, spatial
-from .core import to_8b, to_float, to_hom, homogenize, broadcast_batch, map_range
+from .core import to_8b, to_float, to_hom, homogenize, broadcast_batch, is_np
 from .structures import get_gizmo_coords
 from .gsoup_io import save_image
 
@@ -12,14 +13,20 @@ def alpha_compose(images, bg_color=None):
     :param bg_color: 3 or b x 3
     :return: b x H x W x 3 or H x W x 3
     """
-    if bg_color is None:
-        bg_color = np.array([0., 0., 0.]).astype(np.float32)
     if images.ndim != 3 and images.ndim != 4:
         raise ValueError("image must be 3 or 4 dimensional")
     if images.shape[-1] != 4:
         raise ValueError("image must have 4 channels")
-    if images.dtype != np.float32:
-        images = to_float(images)
+    if is_np(images):
+        if bg_color is None:
+            bg_color = np.array([0., 0., 0.]).astype(np.float32)
+        if images.dtype != np.float32:
+            images = to_float(images)
+    else:
+        if bg_color is None:
+            bg_color = torch.tensor([0., 0., 0.], dtype=images.dtype, device=images.device)
+        if images.dtype != torch.float32:
+            images = to_float(images)
     alpha = images[..., 3:4]
     rgb = images[..., :3]
     return alpha * rgb + (1 - alpha) * bg_color
@@ -49,13 +56,14 @@ def draw_text_on_image(images, text_per_image, fill_white=True):
         rgbs = to_float(rgbs)
     return rgbs
 
-def draw_gizmo_on_image(np_images, w2c, isOpenGL=False, scale=.05):
+def draw_gizmo_on_image(np_images, w2c, opengl=False, scale=.05):
     """
     adds a gizmo to a batch of np images.
     note: will broadcast np_images and w2c against eachother.
     :param np_images: b x H x W x 3
     :param w2c: b x 3 x 4 w2c transforms (opencv conventions)
-    :param isOpenGL: if True, the w2c transforms are assumed to be in OpenGL conventions, else OpenCV conventions
+    :param opengl: if True, the w2c transforms are assumed to be in OpenGL conventions, else OpenCV conventions
+    for opengl, w2c should be a bx4x4 matrix converting from world to *CLIP* space.
     :param scale: scale of the gizmo
     :return: b x H x W x 3
     """
@@ -68,17 +76,14 @@ def draw_gizmo_on_image(np_images, w2c, isOpenGL=False, scale=.05):
     for i, np_image in enumerate(np_images):
         pil_image = Image.fromarray(to_8b(np_image))
         W, H = pil_image.size
-        # W, H = image.shape[1], image.shape[0]
         gizmo_cords, _, _ = get_gizmo_coords(scale)
-        gizmo_hom = to_hom(gizmo_cords)  #  = np.concatenate((gizmo_cords, np.ones_like(gizmo_cords[:, 0:1])), axis=-1)
+        gizmo_hom = to_hom(gizmo_cords)
         verts_clip = (w2c[i] @ gizmo_hom.T).T
-        verts_clip = homogenize(verts_clip) # verts_screen_xy = verts_screen[:, :2] / verts_screen[:, 2:3]
-        if isOpenGL:
-            raise NotImplementedError("OpenGL convention is not supported for now")
-            # if not (np.abs(verts_screen_xy[:, 2]) <= 1).all():
-            #    raise ValueError("OpenGL convention is not followed")
-            # verts_clip = verts_clip[:, :2]
-            # verts_screen = np.array([W, H]) * (verts_clip + 1) / 2
+        verts_clip = homogenize(verts_clip)
+        if opengl:
+            verts_clip = verts_clip[:, :2]
+            verts_screen = np.array([W, H]) * (verts_clip + 1) / 2
+            verts_screen[:, 1] *= -1
         else:
             verts_screen = verts_clip
         desired_loc = np.array([W - 40, H - 40])
@@ -106,7 +111,6 @@ def merge_figures_with_line(img1, img2, lower_intersection=0.6, angle=np.pi/2, l
     else:
         line_color = np.array(line_color, dtype=np.uint8)
     combined = np.ascontiguousarray(np.broadcast_to(line_color, img1.shape))
-    # combined = np.ones_like(img1)*255
     y, x, _ = img1.shape
     yy, xx = np.mgrid[:y, :x]
     img1_positions = (xx-lower_intersection*x)*np.tan(angle)-line_width//2>(yy-y)
@@ -161,6 +165,20 @@ def generate_dot_pattern(height, width, background="black", radius=5, spacing=50
         img.save(str(dst))
     return np.array(img)
 
+def generate_checkerboard(h, w, blocksize):
+    """
+    generates a checkerboard pattern
+    note: if blocksize is not a divisor of w or h, the pattern will have extra "crops" at the edges
+    :param h: height of the image
+    :param w: width of the image
+    :param blocksize: size of the squares
+    :return: (H x W x 1) numpy array (np.bool)
+    """
+    c0, c1 = 0, 1  # color of the squares, for binary these are just 0,1
+    tile = np.array([[c0, c1],[c1, c0]], dtype=np.bool).repeat(blocksize, axis=0).repeat(blocksize, axis=1)[..., None]
+    grid = np.tile(tile, ( h//(2*blocksize)+1, w//(2*blocksize)+1, 1))
+    return grid[:h,:w]
+
 def generate_stripe_pattern(height, width, background="black", direction="vert", thickness=5, spacing=50, dst=None):
     """
     generates an image with colored stripes in a certain direction
@@ -201,6 +219,7 @@ def generate_lollipop_pattern(height, width, background="black", n=15, m=8, dst=
     :param n: number of circles in the pattern
     :param m: number of lines in the pattern
     :param dst: if not None, the image is written to this path
+    :return: (H x W x 3) numpy array (uint8)
     """
     spacing_x = width // (2*n)
     spacing_y = height // (2*n)
@@ -357,25 +376,119 @@ def resize_images_naive(images, H, W, channels_last=True, mode="mean"):
         raise ValueError("images must be square")
     if not channels_last:
         raise NotImplementedError("only channels last is supported")
+    channels_size = images.shape[-1]
     input_size = np.array(images.shape[1:3])
     output_size = np.array([H, W])
     bin_size = input_size // output_size
     if mode == "max":
-        small_images = images.reshape((images.shape[0], output_size[0], bin_size[0], output_size[1], bin_size[1], 3)).max(4).max(2)
+        small_images = images.reshape((images.shape[0], output_size[0], bin_size[0], output_size[1], bin_size[1], channels_size)).max(4).max(2)
     elif mode == "mean":
-        small_images = images.reshape((images.shape[0], output_size[0], bin_size[0], output_size[1], bin_size[1], 3)).mean(4).mean(2)
-        small_images = small_images.astype(np.uint8)
+        small_images = images.reshape((images.shape[0], output_size[0], bin_size[0], output_size[1], bin_size[1], channels_size)).mean(4).mean(2)
+        if images.dtype == np.uint8:
+            small_images = small_images.astype(np.uint8)
     else:
         raise ValueError("mode must be one of 'max', 'mean'")
     return small_images
 
+def pad_image_to_res(images, res_h, res_w, bg_color=None):
+    """
+    pads a batch of numpy images to a specific resolution
+    :param image: numpy image b x h x w x c
+    :param res_h: height of the output image
+    :param res_w: width of the output image
+    :param bg_color: background color c (defaults to black)
+    :return: padded image b x res_h x res_w x c
+    """
+    if bg_color is None:
+        if is_np(images):
+            bg_color = np.zeros(images.shape[-1], dtype=images.dtype)
+        else:
+            bg_color = torch.zeros(images.shape[-1], dtype=images.dtype, device=images.device)
+    if images.ndim != 4:
+        raise ValueError("image must be a 4D array")
+    b, h, w, c = images.shape
+    if h > res_h or w > res_w:
+        raise ValueError("images dimensions is larger than the output resolution")
+    if h == res_h and w == res_w:
+        return images
+    if bg_color.shape[0] != c:
+        raise ValueError("background color must have the same number of channels as the image")
+    bg_color = bg_color[None, None, None, :]
+    if is_np(images):
+        output = np.zeros((b, res_h, res_w, c), dtype=images.dtype)
+    else:
+        output = torch.zeros((b, res_h, res_w, c), dtype=images.dtype, device=images.device)
+    output[:, :, :, :] = bg_color
+    corner_left = (res_w - w) // 2
+    corner_top = (res_h - h) // 2
+    output[:, corner_top:corner_top + h, corner_left:corner_left + w, :] = images
+    return output
+
+def crop_center(images, dst_h, dst_w):
+    """
+    crops a batch of images to a specific resolution, but the crop comes from the center of the image
+    :param image: numpy (or torch) array b x h x w x c
+    :param dst_h: height of the output image
+    :param dst_w: width of the output image
+    :return: cropped image b x dst_h x dst_w x c
+    """
+    if images.ndim != 4:
+        raise ValueError("image must be a 4D array")
+    _, h, w, _ = images.shape
+    if h < dst_h or w < dst_w:
+        raise ValueError("images dimensions is smaller than the output resolution")
+    if h == dst_h and w == dst_w:
+        return images
+    corner_left = (w - dst_w) // 2
+    corner_top = (h - dst_h) // 2
+    output = images[:, corner_top:corner_top + dst_h, corner_left:corner_left + dst_w, :]
+    return output
+
+def mask_regions(images, start_h, end_h, start_w, end_w):
+    """
+    masks a batch of numpy image with black background outside of region of interest (roi)
+    :param image: numpy image b x h x w x c
+    :param start_h: where does the roi height start
+    :param end_h: where does the roi height end
+    :param start_w: where does the roi width start
+    :param end_w: where does the roi width end
+    :return: masked image
+    """
+    if images.ndim != 4:
+        raise ValueError("image must be a 4D array")
+    _, h, w, _ = images.shape
+    if start_h < 0 or start_w < 0 or end_h > h or end_w > w:
+        raise ValueError("Values exceed image resolution")
+    output = np.zeros_like(images)
+    output[:, start_h:end_h, start_w:end_w, :] = images[start_h:end_h, start_w:end_w, :]
+    return output
+
+def adjust_contrast_brightness(img, alpha, beta=None):
+    """
+    adjusts image contrast and brightness using naive gain and bias factors
+    :param img: input image numpy array (n x h x w x 3), float values between 0 and 1
+    :param alpha: gain factor ("contrast") between 0 and inf
+    :param beta: bias factor ("brightness") between -inf and inf (but typically between -1 and 1)
+    :return: the new image
+    """
+    if img.dtype != np.float32:
+        raise ValueError("img must be a float32 numpy array (0-1)")
+    if beta is None:  # if beta is not provided, set to factor of alpha
+        beta = 0.5 - alpha / 2
+    new_img = img * alpha + beta
+    new_img[new_img < 0] = 0
+    new_img[new_img > 1] = 1
+    return new_img.astype(np.uint8)
+
 def change_brightness(input_img, brightness=0):
     """
-    changes brightness of an image
-    :param input_img the numpy image
+    changes brightness of an image or batch of images
+    :param input_img a numpy or torch tensor of float values between 0 and 1 (n x h x w x 3)
     :param brightness a number between -255 to 255 (0=no change)
     :return the new image
     """
+    if input_img.dtype != np.float32 and input_img.dtype != torch.float32:
+        raise ValueError("input_img must be a float32 array (0-1)")
     if brightness != 0:
         if brightness > 0:
             shadow = brightness
@@ -384,25 +497,28 @@ def change_brightness(input_img, brightness=0):
             shadow = 0
             highlight = 255 + brightness
         alpha_b = (highlight - shadow)/255
-        gamma_b = shadow
+        gamma_b = shadow / 255
     else:
         alpha_b = 1
         gamma_b = 0
     return input_img*alpha_b + gamma_b
 
-def change_contrast(input_img, contrast=127):
+def linear_to_srgb(linear):
     """
-    changes brightness of an image
-    :param input_img the numpy image
-    :param brightness a number between -127 to 127 (0=no change)
-    :return the new image
+    converts linear RGB to sRGB, see https://en.wikipedia.org/wiki/SRGB.
+    note: linear is expected to be in the range [0, 1]
     """
-    contrast = map_range(contrast, 0, 254, -127, 127)
-    if contrast != 0:
-        f = float(131 * (contrast + 127)) / (127 * (131 - contrast))
-        alpha_c = f
-        gamma_c = 127*(1-f)
-    else:
-        alpha_c = 1
-        gamma_c = 0
-    return input_img*alpha_c + gamma_c
\ No newline at end of file
+    eps = np.finfo(np.float32).eps
+    srgb0 = 323 / 25 * linear
+    srgb1 = (211 * np.maximum(eps, linear)**(5 / 12) - 11) / 200
+    return np.where(linear <= 0.0031308, srgb0, srgb1)
+
+def srgb_to_linear(srgb):
+    """
+    converts linear RGB to sRGB, see https://en.wikipedia.org/wiki/SRGB.
+    note: srgb is expected to be in the range [0, 1]
+    """
+    eps = np.finfo(np.float32).eps
+    linear0 = 25 / 323 * srgb
+    linear1 = np.maximum(eps, ((200 * srgb + 11) / (211)))**(12 / 5)
+    return np.where(srgb <= 0.04045, linear0, linear1)
\ No newline at end of file
